import logging
from gillespy2.core import gillespyError
from gillespy2.core.gillespy2 import Model, Species, Reaction, Parameter, RateRule, StochMLDocument, import_SBML
from gillespy2.core.gillespySolver import GillesPySolver


_formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
_handler = logging.StreamHandler()
_handler.setFormatter(_formatter)

log = logging.getLogger()
log.setLevel(logging.WARN)
log.addHandler(_handler)
version = '1.1'

__all__ = [s for s in dir() if not s.startswith('_')]
